from collections import OrderedDict

import numpy as np
import torch
import torch.optim as optim

import lfrl.torch.pytorch_util as ptu
from lfrl.util.eval_util import create_stats_ordered_dict
from lfrl.core.rl_algorithms.torch_rl_algorithm import TorchTrainer

class AdaptiveSACTrainer(TorchTrainer):

    def __init__(
            self,
            env,
            policy,
            qf1,
            qf2,
            target_qf1,
            target_qf2,

            discount=0.99,
            reward_scale=1.0,

            policy_lr=1e-3,
            qf_lr=1e-3,
            optimizer_class=optim.Adam,

            soft_target_tau=1e-2,
            target_update_period=1,
            plotter=None,
            render_eval_paths=False,

            use_automatic_entropy_tuning=True,
            target_entropy=None,

            kappa=1, # huber regression threshold

            **kwargs
    ):
        super().__init__()
        self.env = env
        self.policy = policy
        self.qf1 = qf1
        self.qf2 = qf2
        self.target_qf1 = target_qf1
        self.target_qf2 = target_qf2
        self.soft_target_tau = soft_target_tau
        self.target_update_period = target_update_period

        self.use_automatic_entropy_tuning = use_automatic_entropy_tuning
        if self.use_automatic_entropy_tuning:
            if target_entropy:
                self.target_entropy = target_entropy
            else:
                self.target_entropy = -np.prod(self.env.action_space.shape).item()  # heuristic value from Tuomas
            self.log_alpha = ptu.zeros(1, requires_grad=True)
            self.alpha_optimizer = optimizer_class(
                [self.log_alpha],
                lr=policy_lr,
            )

        self.plotter = plotter
        self.render_eval_paths = render_eval_paths

        self.policy_optimizer = optimizer_class(
            self.policy.parameters(),
            lr=policy_lr,
        )
        self.qf1_optimizer = optimizer_class(
            self.qf1.parameters(),
            lr=qf_lr,
        )
        self.qf2_optimizer = optimizer_class(
            self.qf2.parameters(),
            lr=qf_lr,
        )

        self.discount = discount
        self.reward_scale = reward_scale

        self.n_quantiles = self.qf1.n_quantiles
        self.kappa = kappa

        self.eval_statistics = OrderedDict()
        self._n_train_steps_total = 0
        self._need_to_update_eval_statistics = True

        self.policy.latent = np.zeros(1) # + 0.5

    def train_from_torch(self, batch):
        rewards = batch['rewards']
        terminals = batch['terminals']
        obs = batch['observations']
        actions = batch['actions']
        next_obs = batch['next_observations']

        """
        Adaptive risk learning: sample latent risk parameter for update
        """
        latents = ptu.rand(obs.shape[0], 1) # parameter is in (0, 1) - note this is rounded down

        """
        Policy and Alpha Loss
        """
        new_obs_actions, policy_mean, policy_log_std, log_pi, *_ = self.policy(
            obs, latents, reparameterize=True, return_log_prob=True,
        )
        if self.use_automatic_entropy_tuning:
            alpha_loss = -(self.log_alpha * (log_pi + self.target_entropy).detach()).mean()
            self.alpha_optimizer.zero_grad()
            alpha_loss.backward()
            self.alpha_optimizer.step()
            alpha = self.log_alpha.exp()
        else:
            alpha_loss = 0
            alpha = 1

        qf1_weighted_pred = self.qf1(obs, new_obs_actions, latents, risk_parameters=latents)
        qf2_weighted_pred = self.qf2(obs, new_obs_actions, latents, risk_parameters=latents)

        q_new_actions = torch.min(
            qf1_weighted_pred,
            qf2_weighted_pred,
        )
        
        policy_loss = (alpha*log_pi - q_new_actions).mean()

        qf1_weighted_pred = self.qf1(obs, actions, latents, risk_parameters=latents)
        qf2_weighted_pred = self.qf2(obs, actions, latents, risk_parameters=latents)

        """
        QF Loss
        """
        q1_pred = self.qf1.get_quantile_values(obs, actions, latents)
        q1_pred = q1_pred.view(-1, self.n_quantiles, 1)
        q1_pred = q1_pred.repeat(1, 1, self.n_quantiles) # rows

        q2_pred = self.qf2.get_quantile_values(obs, actions, latents)
        q2_pred = q2_pred.view(-1, self.n_quantiles, 1)
        q2_pred = q2_pred.repeat(1, 1, self.n_quantiles) # rows

        new_next_actions, _, _, new_log_pi, *_ = self.policy(
            next_obs, latents, reparameterize=True, return_log_prob=True,
        )

        target_q_values = torch.min(
            self.target_qf1.get_quantile_values(next_obs, new_next_actions, latents),
            self.target_qf2.get_quantile_values(next_obs, new_next_actions, latents),
        ) - alpha * new_log_pi

        q_target = self.reward_scale * rewards + (1. - terminals) * self.discount * target_q_values
        q_target = q_target.detach().view(-1, 1, self.n_quantiles)
        q_target = q_target.repeat(1, self.n_quantiles, 1) # columns

        quantile_thresholds = self.qf1.thresholds.view(1, self.n_quantiles, 1) # this is a constant

        q1_abs_diff = torch.abs(q_target - q1_pred)
        q1_minor_error = (q1_abs_diff <= self.kappa).float().detach() * \
            0.5 * q1_abs_diff ** 2
        q1_major_error = (q1_abs_diff > self.kappa).float().detach() * \
            self.kappa * (q1_abs_diff - 0.5 * self.kappa)
        q1_error = q1_minor_error + q1_major_error
        q1_comparison = (q_target < q1_pred).float().detach()
        q1_punishment = torch.abs(quantile_thresholds - q1_comparison)
        qf1_loss = (q1_punishment * q1_error / self.kappa).sum(dim=2).mean()

        q2_abs_diff = torch.abs(q_target - q2_pred)
        q2_minor_error = (q2_abs_diff <= self.kappa).float().detach() * \
            0.5 * q2_abs_diff ** 2
        q2_major_error = (q2_abs_diff > self.kappa).float().detach() * \
            self.kappa * (q2_abs_diff - 0.5 * self.kappa)
        q2_error = q2_minor_error + q2_major_error
        q2_comparison = (q_target < q2_pred).float().detach()
        q2_punishment = torch.abs(quantile_thresholds - q2_comparison)
        qf2_loss = (q2_punishment * q2_error / self.kappa).sum(dim=2).mean()

        """
        Update networks
        """
        self.qf1_optimizer.zero_grad()
        qf1_loss.backward()
        self.qf1_optimizer.step()

        self.qf2_optimizer.zero_grad()
        qf2_loss.backward()
        self.qf2_optimizer.step()

        self.policy_optimizer.zero_grad()
        policy_loss.backward()
        self.policy_optimizer.step()

        """
        Soft Updates
        """
        if self._n_train_steps_total % self.target_update_period == 0:
            ptu.soft_update_from_to(
                self.qf1, self.target_qf1, self.soft_target_tau
            )
            ptu.soft_update_from_to(
                self.qf2, self.target_qf2, self.soft_target_tau
            )

        """
        Save some statistics for eval
        """
        if self._need_to_update_eval_statistics:
            self._need_to_update_eval_statistics = False
            """
            Eval should set this to None.
            This way, these statistics are only computed for one batch.
            """
            policy_loss = (log_pi - q_new_actions).mean()
            policy_avg_std = torch.exp(policy_log_std).mean()

            self.eval_statistics['QF1 Loss'] = np.mean(ptu.get_numpy(qf1_loss))
            self.eval_statistics['QF2 Loss'] = np.mean(ptu.get_numpy(qf2_loss))
            self.eval_statistics['Policy Loss'] = np.mean(ptu.get_numpy(
                policy_loss
            ))
            self.eval_statistics.update(create_stats_ordered_dict(
                'Q1 Predictions (Weighted)',
                np.mean(ptu.get_numpy(qf1_weighted_pred)),
            ))
            self.eval_statistics.update(create_stats_ordered_dict(
                'Q2 Predictions (Weighted)',
                np.mean(ptu.get_numpy(qf2_weighted_pred)),
            ))
            self.eval_statistics.update(create_stats_ordered_dict(
                'Q1 Predictions (Expected Value)',
                np.mean(ptu.get_numpy(q1_pred)),
            ))
            self.eval_statistics.update(create_stats_ordered_dict(
                'Q2 Predictions (Expected Value)',
                np.mean(ptu.get_numpy(q2_pred)),
            ))
            self.eval_statistics.update(create_stats_ordered_dict(
                'Q Targets',
                ptu.get_numpy(q_target),
            ))
            self.eval_statistics.update(create_stats_ordered_dict(
                'Log Pis',
                ptu.get_numpy(log_pi),
            ))
            self.eval_statistics.update(create_stats_ordered_dict(
                'Policy mu',
                ptu.get_numpy(policy_mean),
            ))
            self.eval_statistics.update(create_stats_ordered_dict(
                'Policy log std',
                ptu.get_numpy(policy_log_std),
            ))
            self.eval_statistics['Policy std'] = np.mean(ptu.get_numpy(policy_avg_std))
            if self.use_automatic_entropy_tuning:
                self.eval_statistics['Alpha'] = alpha.item()
                self.eval_statistics['Alpha Loss'] = alpha_loss.item()
        self._n_train_steps_total += 1

    def get_diagnostics(self):
        return self.eval_statistics

    def end_epoch(self, epoch):
        self._need_to_update_eval_statistics = True

    @property
    def networks(self):
        return [
            self.policy,
            self.qf1,
            self.qf2,
            self.target_qf1,
            self.target_qf2,
        ]

    def get_snapshot(self):
        return dict(
            policy=self.policy,
            qf1=self.qf1,
            qf2=self.qf2,
            target_qf1=self.qf1,
            target_qf2=self.qf2,
        )
